@@ -7,14 +7,64 @@ import numpy as np
 import geopandas as gpd
 from ._vectorized_raster_interpolation import _fast_append_profile_in_gdf
 import warnings
-from scipy.sparse import dok_matrix, diags
+from scipy.sparse import dok_matrix, diags, coo_matrix
 import pandas as pd
+import os
 
-from tobler.util.util import (_check_crs, _nan_check, _inf_check,
-                              _check_presence_of_crs)
+from tobler.util.util import _check_crs, _nan_check, _inf_check, _check_presence_of_crs
 
-def _area_tables_binning(source_df, target_df):
-    """Construct area allocation and source-target correspondence tables using a spatial indexing approach
+
+def _chunk_dfs(geoms_to_chunk, geoms_full, n_jobs):
+    chunk_size = geoms_to_chunk.shape[0] // n_jobs + 1
+    for i in range(n_jobs):
+        start = i * chunk_size
+        yield geoms_to_chunk.iloc[start : start + chunk_size], geoms_full
+
+
+def _index_n_query(geoms1, geoms2):
+    # Pick largest for STRTree, query_bulk the smallest
+    if geoms1.shape[0] > geoms2.shape[0]:
+        large = geoms1
+        small = geoms2
+    else:
+        large = geoms2
+        small = geoms1
+    # Build tree + query
+    qry_polyIDs, tree_polyIDs = large.sindex.query_bulk(small, predicate="intersects")
+    # Remap IDs to global
+    large_global_ids = large.iloc[tree_polyIDs].index.values
+    small_global_ids = small.iloc[qry_polyIDs].index.values
+    # Return always global IDs for geoms1, geoms2
+    if geoms1.shape[0] > geoms2.shape[0]:
+        return np.array([large_global_ids, small_global_ids]).T
+    else:
+        return np.array([small_global_ids, large_global_ids]).T
+
+
+def _chunk_polys(id_pairs, geoms_left, geoms_right, n_jobs):
+    chunk_size = id_pairs.shape[0] // n_jobs + 1
+    for i in range(n_jobs):
+        start = i * chunk_size
+        chunk1 = geoms_left.values.data[id_pairs[start : start + chunk_size, 0]]
+        chunk2 = geoms_right.values.data[id_pairs[start : start + chunk_size, 1]]
+        yield chunk1, chunk2
+
+
+def _intersect_area_on_chunk(geoms1, geoms2):
+    import pygeos
+
+    areas = pygeos.area(pygeos.intersection(geoms1, geoms2))
+    return areas
+
+
+def _area_tables_binning_parallel(source_df, target_df, n_jobs=-1):
+    """Construct area allocation and source-target correspondence tables using
+    a parallel spatial indexing approach
+    ...
+
+    NOTE: currently, the largest df is chunked and the other one is shipped in
+    full to each core; within each process, the spatial index is built for the
+    largest set of geometries, and the other one used for `query_bulk`
 
     Parameters
     ----------
@@ -22,6 +72,94 @@ def _area_tables_binning(source_df, target_df):
         GeoDataFrame containing input data and polygons
     target_df : geopandas.GeoDataFramee
         GeoDataFrame defining the output geometries
+    n_jobs : int
+        [Optional. Default=-1] Number of processes to run in parallel. If -1,
+        this is set to the number of CPUs available
+
+    Returns
+    -------
+    tables : scipy.sparse.dok_matrix
+
+    """
+    from joblib import Parallel, delayed, parallel_backend
+
+    if _check_crs(source_df, target_df):
+        pass
+    else:
+        return None
+    if n_jobs == -1:
+        n_jobs = os.cpu_count()
+
+    df1 = source_df.copy()
+    df2 = target_df.copy()
+
+    # Chunk the largest, ship the smallest in full
+    if df1.shape[0] > df2.shape[1]:
+        to_chunk = df1
+        df_full = df2
+    else:
+        to_chunk = df2
+        df_full = df1
+
+    # Spatial index query
+    ## Reindex on positional IDs
+    to_workers = _chunk_dfs(
+        gpd.GeoSeries(to_chunk.geometry.values, crs=to_chunk.crs),
+        gpd.GeoSeries(df_full.geometry.values, crs=df_full.crs),
+        n_jobs,
+    )
+
+    with parallel_backend("loky", inner_max_num_threads=1):
+        worker_out = Parallel(n_jobs=n_jobs)(
+            delayed(_index_n_query)(*chunk_pair) for chunk_pair in to_workers
+        )
+
+    ids_src, ids_tgt = np.concatenate(worker_out).T
+
+    # Intersection + area calculation
+    chunks_to_intersection = _chunk_polys(
+        np.vstack([ids_src, ids_tgt]).T, df1.geometry, df2.geometry, n_jobs
+    )
+    with parallel_backend("loky", inner_max_num_threads=1):
+        worker_out = Parallel(n_jobs=n_jobs)(
+            delayed(_intersect_area_on_chunk)(*chunk_pair)
+            for chunk_pair in chunks_to_intersection
+        )
+    areas = np.concatenate(worker_out)
+
+    # Build DOK table
+    table = coo_matrix(
+        (
+            areas,
+            (ids_src, ids_tgt),
+        ),
+        shape=(df1.shape[0], df2.shape[0]),
+        dtype=np.float32,
+    )
+    table = table.todok()
+    return table
+
+
+def _area_tables_binning(source_df, target_df, spatial_index):
+    """Construct area allocation and source-target correspondence tables using a spatial indexing approach
+    ...
+
+    NOTE: this currently relies on Geopandas' spatial index machinery
+
+    Parameters
+    ----------
+    source_df : geopandas.GeoDataFrame
+        GeoDataFrame containing input data and polygons
+    target_df : geopandas.GeoDataFramee
+        GeoDataFrame defining the output geometries
+    spatial_index : str
+        Spatial index to use to build the allocation of area from source to
+        target tables. It currently support the following values:
+            - "source": build the spatial index on `source_df`
+            - "target": build the spatial index on `target_df`
+            - "auto": attempts to guess the most efficient alternative.
+              Currently, this option uses the largest table to build the
+              index, and performs a `bulk_query` on the shorter table.
 
     Returns
     -------
@@ -36,88 +174,34 @@ def _area_tables_binning(source_df, target_df):
     df1 = source_df.copy()
     df2 = target_df.copy()
 
-    l1, b1, r1, t1 = df1.total_bounds
-    l2, b2, r2, t2 = df2.total_bounds
-    total_bounds = [min(l1, l2), min(b1, b2), max(r1, r2), max(t1, t2)]
-    n1, k1 = df1.shape
-    n2, k2 = df2.shape
-    numPoly = n1 + n2
-    DELTA = 0.000001
+    # it is generally more performant to use the longer df as spatial index
+    if spatial_index == "auto":
+        if df1.shape[0] > df2.shape[0]:
+            spatial_index = "source"
+        else:
+            spatial_index = "target"
 
-    # constants for bucket sizes
-    BUCK_SM = 8
-    BUCK_LG = 80
-    SHP_SMALL = 1000
-
-    shapebox = total_bounds
-    # bucket size
-    if numPoly < SHP_SMALL:
-        bucketmin = numPoly // BUCK_SM + 2
+    if spatial_index == "source":
+        ids_tgt, ids_src = df1.sindex.query_bulk(df2.geometry, predicate="intersects")
+    elif spatial_index == "target":
+        ids_src, ids_tgt = df2.sindex.query_bulk(df1.geometry, predicate="intersects")
     else:
-        bucketmin = numPoly // BUCK_LG + 2
-        # print 'bucketmin: ', bucketmin
-    # bucket length
-    lengthx = ((shapebox[2] + DELTA) - shapebox[0]) / bucketmin
-    lengthy = ((shapebox[3] + DELTA) - shapebox[1]) / bucketmin
+        raise ValueError(
+            f"'{spatial_index}' is not a valid option. Use 'auto', 'source' or 'target'."
+        )
 
-    # initialize buckets
-    columns1 = [set() for i in range(bucketmin)]
-    rows1 = [set() for i in range(bucketmin)]
-    columns2 = [set() for i in range(bucketmin)]
-    rows2 = [set() for i in range(bucketmin)]
+    areas = df1.geometry.values[ids_src].intersection(df2.geometry.values[ids_tgt]).area
 
-    minbox = shapebox[:2] * 2  # minx,miny,minx,miny
-    binWidth = [lengthx, lengthy] * 2  # lenx,leny,lenx,leny
-    bbcache = {}
-    poly2Column1 = [set() for i in range(n1)]
-    poly2Row1 = [set() for i in range(n1)]
-    poly2Column2 = [set() for i in range(n2)]
-    poly2Row2 = [set() for i in range(n2)]
+    table = coo_matrix(
+        (
+            areas,
+            (ids_src, ids_tgt),
+        ),
+        shape=(df1.shape[0], df2.shape[0]),
+        dtype=np.float32,
+    )
 
-    for i in range(n1):
-        shpObj = df1.geometry.iloc[i]
-        bbcache[i] = shpObj.bounds
-        projBBox = [
-            int((shpObj.bounds[:][j] - minbox[j]) / binWidth[j]) for j in range(4)
-        ]
-        for j in range(projBBox[0], projBBox[2] + 1):
-            columns1[j].add(i)
-            poly2Column1[i].add(j)
-        for j in range(projBBox[1], projBBox[3] + 1):
-            rows1[j].add(i)
-            poly2Row1[i].add(j)
-
-    for i in range(n2):
-        shpObj = df2.geometry.iloc[i]
-        bbcache[i] = shpObj.bounds
-        projBBox = [
-            int((shpObj.bounds[:][j] - minbox[j]) / binWidth[j]) for j in range(4)
-        ]
-        for j in range(projBBox[0], projBBox[2] + 1):
-            columns2[j].add(i)
-            poly2Column2[i].add(j)
-        for j in range(projBBox[1], projBBox[3] + 1):
-            rows2[j].add(i)
-            poly2Row2[i].add(j)
-
-    table = dok_matrix((n1, n2), dtype=np.float32)
-
-    for polyId in range(n1):
-        idRows = poly2Row1[polyId]
-        idCols = poly2Column1[polyId]
-        rowNeighbors = set()
-        colNeighbors = set()
-        for row in idRows:
-            rowNeighbors = rowNeighbors.union(rows2[row])
-        for col in idCols:
-            colNeighbors = colNeighbors.union(columns2[col])
-        neighbors = rowNeighbors.intersection(colNeighbors)
-        for neighbor in neighbors:
-            if df1.geometry.iloc[polyId].intersects(df2.geometry.iloc[neighbor]):
-                intersection = df1.geometry.iloc[polyId].intersection(
-                    df2.geometry.iloc[neighbor]
-                )
-                table[polyId, neighbor] = intersection.area
+    table = table.todok()
 
     return table
 
@@ -187,6 +271,8 @@ def _area_interpolate_binning(
     intensive_variables=None,
     table=None,
     allocate_total=True,
+    spatial_index="auto",
+    n_jobs=1,
 ):
     """
     Area interpolation for extensive and intensive variables.
@@ -196,15 +282,34 @@ def _area_interpolate_binning(
     source_df : geopandas.GeoDataFrame
     target_df : geopandas.GeoDataFrame
     extensive_variables : list
-        columns in dataframes for extensive variables
+        [Optional. Default=None] Columns in dataframes for extensive variables
     intensive_variables : list
-        columns in dataframes for intensive variables
+        [Optional. Default=None] Columns in dataframes for intensive variables
     table : scipy.sparse.dok_matrix
+        [Optional. Default=None] Area allocation source-target correspondence
+        table. If not provided, it will be built from `source_df` and
+        `target_df` using `tobler.area_interpolate._area_tables_binning`
     allocate_total : boolean
-        True if total value of source area should be allocated.
-        False if denominator is area of i. Note that the two cases
-        would be identical when the area of the source polygon is
-        exhausted by intersections. See Notes for more details.
+        [Optional. Default=True] True if total value of source area should be
+        allocated. False if denominator is area of i. Note that the two cases
+        would be identical when the area of the source polygon is exhausted by
+        intersections. See Notes for more details.
+    spatial_index : str
+        [Optional. Default="auto"] Spatial index to use to build the
+        allocation of area from source to target tables. It currently support
+        the following values:
+            - "source": build the spatial index on `source_df`
+            - "target": build the spatial index on `target_df`
+            - "auto": attempts to guess the most efficient alternative.
+              Currently, this option uses the largest table to build the
+              index, and performs a `bulk_query` on the shorter table.
+        This argument is ignored if n_jobs>1 (or n_jobs=-1).
+    n_jobs : int
+        [Optional. Default=1] Number of processes to run in parallel to
+        generate the area allocation. If -1, this is set to the number of CPUs
+        available. If `table` is passed, this is ignored.
+        NOTE: as of Jan'21 multi-core functionality requires master versions
+        of `pygeos` and `geopandas`.
 
     Returns
     -------
@@ -249,7 +354,10 @@ def _area_interpolate_binning(
         return None
 
     if table is None:
-        table = _area_tables_binning(source_df, target_df)
+        if n_jobs == 1:
+            table = _area_tables_binning(source_df, target_df, spatial_index)
+        else:
+            table = _area_tables_binning_parallel(source_df, target_df, n_jobs=n_jobs)
 
     den = source_df[source_df.geometry.name].area.values
     if allocate_total:
